/*
 * playitsam: Play, edit & save a TiVo video stream from stdin, file or MFS.
 *
 * tyserver.c: Remote tystream server. $Revision: 1.11 $
 *
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 *
 * Released under the Gnu General Public License version 2
 */

#define USING_MMAP
#include <sys/types.h>
#ifdef USING_MMAP
# include <sys/stat.h>
# include <sys/mman.h>
#endif
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <stdio.h>
#include <dirent.h>


char *streamdir;
#define CHUNK_SIZE (256*512)
char chunk[CHUNK_SIZE];

void list_directory(int sockfd)
{
 DIR *d;
 struct dirent *dp;
 char *errmsg= "400 Cannot open stream directory\r\n";
 char *okmsg= "200 Ok, here is the list of available streams\r\n";

 d= opendir(streamdir);
 if (d==NULL) {
   write(sockfd, errmsg, strlen(errmsg));
   close(sockfd); exit(1);
 }
 write(sockfd, okmsg, strlen(okmsg));

 while ((dp = readdir(d)) != NULL) {
   if ((strlen(dp->d_name)==1) && !strcmp(dp->d_name, ".")) continue;
   if ((strlen(dp->d_name)==2) && !strcmp(dp->d_name, "..")) continue;
   write(sockfd, dp->d_name, strlen(dp->d_name)); write(sockfd, "\r\n", 2);
 }
 (void)closedir(d);
 close(sockfd);
 exit(0);
}


void play_file(int sockfd, char *filename)
{
 int err;
 size_t numread;
#ifdef USING_MMAP
 int infd;
 struct stat S;
 void *mapstart, *bptr;
 size_t maplen, numleft;
#else
 FILE *infile;
#endif
 
 char *errmsg1= "400 Cannot open stream directory\r\n";
 char *errmsg2= "400 Cannot open requested stream\r\n";
 char *okmsg= "200 Ok, here is the stream\r\n";

 err=chdir(streamdir);
 if (err==-1) {
   write(sockfd, errmsg1, strlen(errmsg1));
   close(sockfd); exit(1);
 }

#ifdef USING_MMAP
  infd= open(filename, O_RDONLY);
 if (infd==-1) {
   write(sockfd, errmsg2, strlen(errmsg2));
   close(sockfd); exit(1);
 }

 err=stat(filename, &S);
 if (err==-1) {
   write(sockfd, errmsg2, strlen(errmsg2));
   close(sockfd); exit(1);
 }

 mapstart=NULL; maplen=S.st_size;
 bptr=mmap(mapstart, maplen,  PROT_READ , MAP_PRIVATE, infd, 0);
 if (bptr==MAP_FAILED) {
   write(sockfd, errmsg2, strlen(errmsg2));
   close(sockfd); exit(1);
 }

 write(sockfd, okmsg, strlen(okmsg));

 numleft= maplen;
 while (numleft) {
   numread= write(sockfd, bptr, 1460);
   if (numread<=0) break;
   bptr+= numread;
 }
 close(infd);
#else
 infile= fopen(filename, "r");
 if (infile==NULL) {
   write(sockfd, errmsg2, strlen(errmsg2));
   close(sockfd); exit(1);
 }

 write(sockfd, okmsg, strlen(okmsg));

 while (1) {
   numread = fread(chunk, 1, CHUNK_SIZE, infile);
   if (numread < 1) break;
   write(sockfd, chunk, numread);
 }
 fclose(infile);
#endif

 close(sockfd); exit(0);
}

void save_file(int sockfd, char *filename)
{
 int err;
 int numread;
 FILE *outfile;
 
 char *errmsg1= "400 Cannot open stream directory\r\n";
 char *errmsg2= "400 Cannot save requested stream\r\n";
 char *okmsg= "200 Ok, send me the stream\r\n";

 err=chdir(streamdir);
 if (err==-1) {
   write(sockfd, errmsg1, strlen(errmsg1));
   close(sockfd); exit(1);
 }

 outfile= fopen(filename, "w");
 if (outfile==NULL) {
   write(sockfd, errmsg2, strlen(errmsg2));
   close(sockfd); exit(1);
 }

 write(sockfd, okmsg, strlen(okmsg));

 while (1) {
   numread = read(sockfd, chunk, CHUNK_SIZE);
   if (numread < 0) { close(sockfd); fclose(outfile); exit(1); }
   if (numread == 0) break;
   fwrite(chunk, 1, numread, outfile);
 }
 close(sockfd); fclose(outfile); exit(0);
}


void handle_request(int sockfd)
{
#define BUFSIZE 1024
  char cmd[BUFSIZE];
  int rval;
  int i;

  /* Commands are single lines ending in \r\n. Only 3 commands:
   *	LIST\r\n	to list streams in this directory
   *    PLAY ffff\r\n   to play the file ffff. No slashes in ffff
   *    SAVE ffff\r\n   to save the file ffff. No slashes in ffff
   *
   * Server always returns 1 response line, either
   *	200 Ok\r\n
   * 	400 Error\r\n
   * Client should read 3 digits & ignore the rest.
   *
   * LIST returns a set of ASCII lines terminated in \r\n. Each line
   * contains a filename that can be used with PLAY.
   *
   * PLAY returns a tystream until it ends or the connection is broken.
   *
   * SAVE sucks a tystream from the client until it ends or the connection
   * is broken.
   */

   /* Read a command from the socket */
   /* Ok, so I should read a whole buffer in one hit, who cares! */
   for (i=0;i<BUFSIZE-2;i++) {
     rval= read(sockfd, &cmd[i], 1);
     if ((rval==-1) || (cmd[i]=='\n')) break;
   }
   cmd[++i]='\0';		/* Null-terminate the command */

   if (!strncmp(cmd, "LIST", 4)) list_directory(sockfd);

   /* Walk the command and remove \r and \n */
   for (i=0;i<BUFSIZE;i++) {
     if ((cmd[i]=='\r') || (cmd[i]=='\n')) cmd[i]='\0';
   }
   if (!strncmp(cmd, "PLAY", 4)) play_file(sockfd, &cmd[5]);
   if (!strncmp(cmd, "SAVE", 4)) save_file(sockfd, &cmd[5]);
   close(sockfd);
   exit(0);
}

void usage(void)
{
 printf("Usage: tyserver streams_directory\n");
 exit(1);
}

int main(int argc, char *argv[])
{
  int sock, length;
  struct sockaddr_in server;
  int pid, msgsock;
#if 0
  int nodelay = 1;
#endif

  if (argc!=2) usage();
  streamdir=(char *)strdup(argv[1]);
  if (streamdir==NULL) {
    printf("Wow, can't even strdup these days!\n"); exit(1);
  }

  /* Create socket */
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    perror("opening stream socket");
    exit(1);
  }
  /* Name socket using wildcards */
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;

				/* Why 1150? T looks like 1, i looks like 1
				 * V is 5 in roman numerals, o looks like 0
				 */
  server.sin_port = ntohs(1150);	
  if (bind(sock, (struct sockaddr *)&server, sizeof(server))) {
    perror("binding stream socket");
    exit(1);
  }
  /* Find out assigned port number and print it out */
      length = sizeof(server);
  if (getsockname(sock, (struct sockaddr *)&server, &length)) {
    perror("getting socket name");
    exit(1);
  }
      printf("Socket has port #%d\n", ntohs(server.sin_port));

  /* Start accepting connections */
  listen(sock, 5);

  while (1) {
    msgsock = accept(sock, 0, 0);
    if (msgsock == -1) perror("accept");
#if 0
    setsockopt(msgsock, IPPROTO_TCP, TCP_NODELAY, &nodelay, sizeof(int));
#endif

#define FORK
#ifdef FORK
    pid=fork();
    if (pid==0) handle_request(msgsock);
#else
    handle_request(msgsock);
#endif
  }
}
